#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include "cx88_video_ioctl.h"
#include "tuner_devnode_device.h"
#include "lgh064f.h"
#include "tda9887.h"

void capture(int fd, void *bufspace, cx88_video_format &fmt, cx88_video_config &config)
{
   cx88_video_capture capture;
   capture.in_input_source = CX88_VIDEO_INPUT_MUX1;
   capture.in_input_type = CX88_VIDEO_INPUT_COMPOSITE;
   capture.in_input_options = 0;
   capture.in_lock_timeout_ms = 1000;
   capture.in_lock_delay_ms = 100;
   capture.in_no_frame_intrs = 0;
   printf("Beginning capture\n");
   int error = ioctl(fd, CX88_VIDEO_CAPTURE, &capture);
   if (error)
   {
      printf("Error during capture ioctl\n");
      return;
   }
   char filename[32];
   for (int i = 0; i < config.in_num_frames; ++i)
   {
      cx88_video_wait wait;
      wait.in_timeout_ms = 0;
      if ((i == 0) || (wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available <= i))
      {
         wait.in_channel_mask = CX88_VIDEO_FIELD_EVEN(CX88_VIDEO_CHANNEL_Y);
         error = ioctl(fd, CX88_VIDEO_WAIT, &wait);
         if (error)
         {
            printf("Error during wait ioctl\n");
            ioctl(fd, CX88_VIDEO_STOP);
            return;
         }
         if ((wait.out_status & CX88_VIDEO_BUS_ERROR) == CX88_VIDEO_BUS_ERROR)
         {
            printf("Fatal error: PCI bus error\n");
         }
         if ((wait.out_status & CX88_VIDEO_SYNC_ERROR) == CX88_VIDEO_SYNC_ERROR)
         {
            printf("Fatal error[%d]: DMA sync error\n", i);  
         }
         if ((wait.out_status & CX88_VIDEO_DMA_PROGRAM) == CX88_VIDEO_DMA_PROGRAM)
         {
            printf("Fatal error[%d]: DMA program error\n", i);
         }
         if ((wait.out_status & CX88_VIDEO_WAIT_ABORTED) == CX88_VIDEO_WAIT_ABORTED)
         {
            printf("Fatal error: Wait aborted\n");
         }
         if ((wait.out_status & CX88_VIDEO_CAP_TIMEOUT) == CX88_VIDEO_CAP_TIMEOUT)
         {
            printf("Nonfatal error: Capture timed out\n");
         }
         if (wait.out_status & CX88_VIDEO_CAP_FATAL)
         {
            ioctl(fd, CX88_VIDEO_STOP);
            return;
         }
      }
      if ((wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available <= i) ||
          (wait.channels[CX88_VIDEO_CHANNEL_Y].out_odd_frame_available <= i))
      {
         printf("Warning: No frame available %d %d\n",
            wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available,
            wait.channels[CX88_VIDEO_CHANNEL_Y].out_odd_frame_available);
         /*ioctl(fd, CX88_VIDEO_STOP);
         return;*/
      }
      printf("VBI frames: %d, %d\n",
         wait.channels[CX88_VIDEO_CHANNEL_VBI].out_odd_frame_available,
         wait.channels[CX88_VIDEO_CHANNEL_VBI].out_even_frame_available);
      printf("U frames: %d, %d\n",
         wait.channels[CX88_VIDEO_CHANNEL_U].out_odd_frame_available,
         wait.channels[CX88_VIDEO_CHANNEL_U].out_even_frame_available);
      printf("V frames: %d, %d\n",
         wait.channels[CX88_VIDEO_CHANNEL_V].out_odd_frame_available,
         wait.channels[CX88_VIDEO_CHANNEL_V].out_even_frame_available);
      sprintf(filename, "out%d.ppm", i);
      FILE *outfile = fopen(filename, "w");
      if (outfile == NULL)
      {
         printf("Error opening video output file\n");
         ioctl(fd, CX88_VIDEO_STOP);
         return;  
      }
      uint32_t linesize = fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].line_size + config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_x_header + config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_x_trailer;
      uint32_t numlines = fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].num_lines + config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_y_header + config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_y_trailer;
      fprintf(outfile, "P6\n%d %d\n255\n", linesize, numlines);
      uint32_t bufsize = linesize * numlines;
      uint8_t *ybuf = reinterpret_cast <uint8_t*> 
         (config.bufconfig[CX88_VIDEO_CHANNEL_Y].u.in_userbuf.buffers[i]);
      uint8_t *ubuf = reinterpret_cast <uint8_t*>  
         (config.bufconfig[CX88_VIDEO_CHANNEL_U].u.in_userbuf.buffers[i]);
      uint8_t *vbuf = reinterpret_cast <uint8_t*>  
         (config.bufconfig[CX88_VIDEO_CHANNEL_V].u.in_userbuf.buffers[i]);
      for (size_t j = 0; j < bufsize; ++j)
      {
         uint8_t pixel[3];
         int8_t u = ubuf[j / 2] - 128;
         int8_t v = vbuf[j / 2] - 128;
         double r = 1.370705 * v + ybuf[j];
         double g = (double)(ybuf[j]) - (0.337633 * u) - (0.698001 * v);
         double b = 1.732446 * u + ybuf[j];
         if (r < 0) r = 0;
         if (g < 0) g = 0;
         if (b < 0) b = 0;
         if (r > 255) r = 255;
         if (g > 255) g = 255;
         if (b > 255) b = 255;
         pixel[0] = (uint8_t) r;
         pixel[1] = (uint8_t) g;
         pixel[2] = (uint8_t) b;
         fwrite(pixel, 3, 1, outfile);
      }
      for (size_t j = 0; j < bufsize; j += 3)
      {
         uint8_t pixel[3];
         pixel[0] = ybuf[j + 2];
         pixel[1] = ybuf[j + 1];
         pixel[2] = ybuf[j];
         fwrite(pixel, 3, 1, outfile);
      }
      fclose(outfile);
      if ((wait.channels[CX88_VIDEO_CHANNEL_VBI].out_even_frame_available <= i) ||
          (wait.channels[CX88_VIDEO_CHANNEL_VBI].out_odd_frame_available <= i))
      {
         printf("Warning: no VBI frame available\n");
         continue;
      }
      bufsize = fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].line_size * fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines;
      char *vbibuf = reinterpret_cast <char*> 
         (config.bufconfig[CX88_VIDEO_CHANNEL_VBI].u.in_userbuf.buffers[i]);
      sprintf(filename, "out%d.vbi", i);
      FILE *vbifile = fopen(filename, "w");
      if (vbifile == NULL)
      {
         printf("Warning: unable to open VBI output file\n");
         continue;
      }
      fwrite(vbibuf, bufsize, 1, vbifile);
      fclose(vbifile);
   }
   ioctl(fd, CX88_VIDEO_STOP);
}

void tune(char *tuner_devnode, int fd, int buffers, void *bufspace, cx88_video_format &fmt, cx88_video_config &config)
{
   int error = 0;
   avb_channel channel;
   channel.video_format = AVB_VIDEO_FMT_NTSC_M;
   channel.audio_format = AVB_AUDIO_FMT_AUTO;
   channel.frequency_hz = 177000000;
   channel.bandwidth_hz = 6000000;
   tuner_config conf;
   error = conf.load_file("tuner_config");
   if (error)
   {
      printf("Error loading tuner config file\n");
      return;
   }
   tuner_devnode_device tuner_dev(conf, tuner_devnode, error);
   if (error)
   {
      printf("Error creating tuner devnode\n");
      return;
   }
   lgh064f tuner_driver(conf, tuner_dev);
   error = tuner_driver.set_channel(channel);
   if (error)
   {
      printf("Error setting channel on tuner driver\n");
      return;  
   }
   printf("Ready to start tuner\n");
   error = tuner_driver.start(10000);
   if (error)
   {
      printf("Error starting tuner\n");
      return;
   }
   sleep(1);
   capture(fd, bufspace, fmt, config);
}

int main(void)
{
   cx88_video_format fmt;
   fmt.in_video_format = CX88_VIDEO_FMT_NTSCM;
   fmt.in_pixel_format = CX88_VIDEO_YCC422;
   fmt.in_rows = 480;
   fmt.in_cols = 640;
   fmt.in_x_offset = 0;
   fmt.in_x_size = 640;
   fmt.in_y_offset = 0;
   fmt.in_y_size = 480;
   fmt.in_enable_vbi = 1;
   fmt.in_interlace = 1;
   fmt.in_even_field.swap = 0;
   fmt.in_even_field.enable = 1;
   fmt.in_odd_field.swap = 0;
   fmt.in_odd_field.enable = 1;
   cx88_video_config config;
   uint32_t id;
   int fd = open("/dev/cx88video1", O_RDWR);
   int error = ioctl(fd, CX88_COMMON_CARD_ID, &id);
   if (error)
   {
      printf("Error during card ID ioctl\n");
      return 1;
   }
   printf("Card ID = 0x%x\n", id);
   error = ioctl(fd, CX88_VIDEO_FORMAT, &fmt);
   if (error)
   {  
      printf("Error during format ioctl: %s\n", strerror(error));
      return 1;
   }
   cx88_common_i2c_config i2config;
   i2config.in_bus_number = 0;
   sprintf(i2config.devices[0].in_driver, "iicdev");
   i2config.devices[0].in_speed = CX88_I2C_SPEED_100KHZ;
   i2config.devices[0].in_addr = 0x61;
   i2config.in_num_i2c_devices = 1;
   error = ioctl(fd, CX88_COMMON_I2C_CONFIG, &i2config);
   if (error)
   {
      printf("Error during I2C config ioctl\n");
      return 1;  
   }
   char tuner_devnode[16];
   sprintf(tuner_devnode, "/dev/iicdev%d", i2config.devices[0].out_instance);
   printf("Tuner: %s\n", tuner_devnode);
   size_t xheader = 120;
   size_t xtrailer = 240;
   size_t yheader = 33;
   size_t ytrailer = 56;
   uint8_t num_frames = 8;
   size_t y_size = (xheader + xtrailer + fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].line_size) * 
      (yheader + ytrailer + fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].num_lines);
   size_t u_size = (((xheader + xtrailer) / 2) + fmt.out_geoms[CX88_VIDEO_CHANNEL_U].line_size)
      * (yheader + ytrailer + fmt.out_geoms[CX88_VIDEO_CHANNEL_U].num_lines);
   size_t v_size = (((xheader + xtrailer) / 2) + fmt.out_geoms[CX88_VIDEO_CHANNEL_V].line_size) 
      * (yheader + ytrailer + fmt.out_geoms[CX88_VIDEO_CHANNEL_V].num_lines);
   size_t vbi_size = fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].line_size * fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines;
   void *buf = malloc(num_frames * (y_size + u_size + v_size + vbi_size));
   if (buf == NULL)
   {
      printf("Error allocating buffer\n");
      return 1;
   }
   for (uint8_t frames = num_frames; frames >= num_frames - 2; frames -= 2)
   {
      sleep(10);
      config.in_num_frames = frames;
      config.in_one_shot = 1;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_buftype = CX88_VIDEO_BUF_USER;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_buftype = CX88_VIDEO_BUF_USER;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_buftype = CX88_VIDEO_BUF_USER;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_buftype = CX88_VIDEO_BUF_USER;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_y_header = yheader;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_y_trailer = ytrailer;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_x_header = xheader;
      config.bufconfig[CX88_VIDEO_CHANNEL_Y].in_x_trailer = xtrailer;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_y_header = yheader;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_y_trailer = ytrailer;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_x_header = xheader / 2;
      config.bufconfig[CX88_VIDEO_CHANNEL_U].in_x_trailer = xtrailer / 2;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_y_header =  yheader;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_y_trailer = ytrailer;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_x_header = xheader / 2;
      config.bufconfig[CX88_VIDEO_CHANNEL_V].in_x_trailer = xtrailer / 2;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_x_header = 0;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_x_trailer = 0;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_y_header = 0;
      config.bufconfig[CX88_VIDEO_CHANNEL_VBI].in_y_trailer = 0;
      for (uint8_t i = 0; i < frames; ++i)
      {
         config.bufconfig[CX88_VIDEO_CHANNEL_Y].u.in_userbuf.buffers[i] = (char*)buf + (i * y_size);
         config.bufconfig[CX88_VIDEO_CHANNEL_U].u.in_userbuf.buffers[i] = (char*)buf + (num_frames * y_size) + (i * u_size);
         config.bufconfig[CX88_VIDEO_CHANNEL_V].u.in_userbuf.buffers[i] = (char*)buf + (num_frames * (y_size + u_size)) + (i * v_size);
         config.bufconfig[CX88_VIDEO_CHANNEL_VBI].u.in_userbuf.buffers[i] = (char*)buf + (num_frames * (y_size + u_size + v_size)) + (i * vbi_size);
      }
      printf("Y Geometry: %d x %d\n", fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].line_size, fmt.out_geoms[CX88_VIDEO_CHANNEL_Y].num_lines);
      printf("U Geometry: %d x %d\n", fmt.out_geoms[CX88_VIDEO_CHANNEL_U].line_size, fmt.out_geoms[CX88_VIDEO_CHANNEL_U].num_lines);
      printf("V Geometry: %d x %d\n", fmt.out_geoms[CX88_VIDEO_CHANNEL_V].line_size, fmt.out_geoms[CX88_VIDEO_CHANNEL_V].num_lines);
      printf("VBI Geometry: %d x %d\n", fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].line_size, fmt.out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines);
      error = ioctl(fd, CX88_VIDEO_CONFIG, &config);
      if (error)
      {
         printf("Error during config ioctl: %s\n", strerror(error));
         return 1;
      }
      tune(tuner_devnode, fd, 400, buf, fmt, config);
   }
   printf("Ready to free buffer\n");
   close(fd);
   free(buf);
   printf("Freed buffer\n");
   return 0;
}

